#include <cstdio>
#include <vector>
#include <argh.hh>

using namespace std;

#define FOURCC 0x52474218

static FILE      *fout;
static unsigned frameus = 1000000 / 25;
static vector<string> files;

static unsigned width, height;

static void dwf(unsigned dw)
{
    fputc((unsigned char)((dw      ) & 255), fout);
    fputc((unsigned char)((dw >> 8 ) & 255), fout);
    fputc((unsigned char)((dw >> 16) & 255), fout);
    fputc((unsigned char)((dw >> 24) & 255), fout);
}
static void wf(unsigned short dw)
{
    fputc((unsigned char)((dw      ) & 255), fout);
    fputc((unsigned char)((dw >> 8 ) & 255), fout);
}

class SizeHere
{
    unsigned pos;
public:
    SizeHere()
    {
        pos = ftell(fout);
        fwrite("----", 4, 1, fout);
    }
    ~SizeHere()
    {
        unsigned now = ftell(fout);
        fseek(fout, pos, SEEK_SET);
        dwf(now - pos - 4);
        fseek(fout, now, SEEK_SET);
    }
};

static void CalcSize()
{
    width = 0;
    height = 0;
    if(!files.size())return;
    FILE *fp = fopen(files[0].c_str(), "rb");
    if(!fp)return;
    fseek(fp, 12, SEEK_SET);
    width = fgetc(fp);  width |= fgetc(fp) << 8;
    height = fgetc(fp); height |= fgetc(fp) << 8;
    fclose(fp);
}
static void WriteFile(const string &fn)
{
    FILE *fp = fopen(fn.c_str(), "rb");
    if(!fp)
    {
        perror(fn.c_str());
        return;
    }
    unsigned datapos = fgetc(fp) + 18;
    fseek(fp, 12, SEEK_SET);
    width = fgetc(fp);  width |= fgetc(fp) << 8;
    height = fgetc(fp); height |= fgetc(fp) << 8;
    
    char *Buf = new char[width*3];
    for(unsigned y=height; y-->0; )
    {
        fseek(fp, datapos + y*width*3, SEEK_SET);
        fread(Buf, 3, width, fp);
        fwrite(Buf, 3, width, fout);
        /*
        for(unsigned x=width; x-->0; )
            fwrite(&Buf[x*3], 1, 3, fout);
        */
    }
    delete[] Buf;
    
    fclose(fp);
}
static void WriteSTRH()
{
    /* TSS EI OLE VIKAA! */
    fwrite("strh", 4, 1, fout);
    SizeHere tmp;
    fwrite("vids", 4, 1, fout);
    dwf(FOURCC);// FOURCC
    dwf(0);     // flags
    wf(0);      // priority
    wf(0);      // language
    dwf(0);     // initial frames
    dwf(1000000);                  // scale
    dwf((unsigned)(1e12/frameus)); // rate  (put scale*fps here)
    dwf(0);     // start
    dwf(0);     // scale*framecount
    dwf(1048576);// suggested buffer size
    dwf(10000); // quality
    dwf(width*height*3);
    wf(0);       // left
    wf(0);       // top
    wf(width);   // right
    wf(height);  // bottom
}
static void WriteSTRF()
{
    fwrite("strf", 4, 1, fout);
    SizeHere tmp;
    dwf(40);    // size - ????
    dwf(width); // width
    dwf(height);// height
    wf(1);      // planes
    wf(24);     // depth
    /* compression type */
    dwf(FOURCC);     // FOURCC
    dwf(width*height*3); // width*height*3
    dwf(0);     // ?
    dwf(0);     // ?
    dwf(0);     // ?
    dwf(0);     // ?
}
static void WriteAVIH()
{
    fwrite("avih", 4, 1, fout);
    SizeHere tmp;
    dwf(frameus);   // microseconds per frame
    dwf((unsigned)(width*height*3e6/frameus)); // max bytes per sec
    dwf(0);         // padding (0)
    dwf(0);         // flags
    dwf(files.size());// total frames
    dwf(0);         // initial frames
    dwf(1);         // streamcount (1, only one: video)
    dwf(1048576);   // suggested buffer size
    dwf(width);     // width
    dwf(height);    // height
    dwf(0);dwf(0);dwf(0);dwf(0); // reserved
}
static void WriteSTRL()
{
    fwrite("LIST", 4, 1, fout);
    SizeHere tmp;
    fwrite("strl", 4, 1, fout);
    WriteSTRH();
    WriteSTRF();
}
static void WriteHDRL()
{
    fwrite("LIST", 4, 1, fout);
    SizeHere tmp;
    fwrite("hdrl", 4, 1, fout);
    WriteAVIH();
    WriteSTRL();
}
static void WriteMOVI()
{
    fwrite("LIST", 4, 1, fout);
    SizeHere tmp;
    fwrite("movi", 4, 1, fout);
    
    for(unsigned a=0; a<files.size(); ++a)
    {
        char Buf[5] = "00dc"; // 00 = stream index
        fwrite(Buf, 4, 1, fout);
        SizeHere tmp;
        WriteFile(files[a]);
    }
}

static void WriteAVI()
{
    fwrite("RIFF", 4, 1, fout);
    SizeHere tmp;
    fwrite("AVI ", 4, 1, fout);
    WriteHDRL();
    WriteMOVI();
}

int main(int argc, const char *const *argv)
{
    string outfile;
    
    ParamHandler Argh;
    Argh.AddLong("help",    'h').SetBool().SetDesc("This help.");
    Argh.AddLong("version", 'V').SetBool().SetDesc("Displays version information.");
    Argh.AddLong("outfile", 'o').SetString(1).SetDesc("Specify the output file.");
    Argh.AddLong("fps",     'f').SetFloat(23e-5, 1e6).SetDesc("Frame rate (default: 25).");
    
    Argh.StartParse(argc, argv);
    for(;;)
    {
        int c = Argh.GetParam();
        if(c == -1)break;
        switch(c)
        {
            case 'V': printf("%s\n", VERSION); return 0;
            case 'h':
                printf(
                    "This is animmaker-"VERSION"\n"
                    "Copyright (C) 1992,2002 Bisqwit (http://iki.fi/bisqwit/)\n"
                    "\nUsage: animmaker [<option> [<...>]] <file> [<...>]\n"
                    "\nThis software assumes the files are uncompressed rgb24 tga files.\n"
                    "\nOptions:\n");
                Argh.ListOptions();
                printf("\nNo warranty whatsoever.\n");
                return 0;
            case 'o':
                outfile = Argh.GetString();
                break;
            case 'f':
                frameus = (unsigned)(1e6 / Argh.GetFloat());
                break;
            default:
                files.push_back(Argh.GetString());
        }
    }
    if(!Argh.ok())return -1;
    if(!files.size())
    {
        fprintf(stderr, "Error: At least one file must be specified. Try --help.\n");
        return -1;
    }
    
    if(!outfile.size() || outfile == "-")
    {
        fout = stdout;
    }
    else
        fout = fopen(outfile.c_str(), "wb");
    if(!fout)
    {
        perror(outfile.c_str());
        return -1;
    }
    
    CalcSize();
    WriteAVI();
    
    fclose(fout);
    
    return 0;
}
